import os
import json
import shlex
import warnings
import subprocess
import itertools
import operator
from glob import glob
from functools import reduce
from copy import deepcopy
from typing import List, Dict, Any, Optional, Callable, cast

import numpy as np
from scipy.interpolate import interp1d, RegularGridInterpolator
from inference_interface import toyfiles_to_numpy

from alea.runner import Runner
from alea.submitter import Submitter
from alea.utils import (
    load_json,
    asymptotic_critical_value,
    deterministic_hash,
    search_filename_pattern,
    get_metadata,
)


class SubmitterLocal(Submitter):
    """Submitter for local machine."""

    def __init__(self, *args, **kwargs):
        """Initialize the SubmitterLocal class."""
        self.local_configurations = kwargs.get("local_configurations", {})
        self.template_path = self.local_configurations.pop("template_path", None)
        self.combine_n_jobs = self.local_configurations.pop("combine_n_jobs", 1)
        self.first_i_batch = kwargs.pop("i_batch", 0)
        super().__init__(*args, **kwargs)

    @staticmethod
    def initialized_runner(script: str, pop_limit_threshold: bool = False):
        """Initialize a Runner from a script.

        Args:
            script: the script to initialize the Runner
            pop_limit_threshold: whether to pop the limit_threshold from the
                statistical_model_args, this is needed for the NeymanConstructor,
                to initialize runner when the limit_threshold does not exist yet.

        """
        kwargs = Submitter.runner_kwargs_from_script(shlex.split(script)[2:])
        if pop_limit_threshold:
            kwargs["statistical_model_args"].pop("limit_threshold", None)
        runner = Runner(**kwargs)
        return runner

    def submit(self, *args, **kwargs):
        """Run job in subprocess locally.

        If debug is True, only return the first instance of Runner.

        """
        for _, (script, _) in enumerate(self.combined_tickets_generator()):
            if self.debug:
                if kwargs:
                    print(f"{' KWARGS ':#^80}")
                    print(kwargs)
                    _, _, annotations = Runner.runner_arguments()
                    runner_kwargs = Submitter.runner_kwargs_from_script(shlex.split(script)[2:])
                    runner_kwargs.update(kwargs)
                    script = Submitter.script_from_runner_kwargs(annotations, runner_kwargs)
                    script = f"python3 {self.run_toymc} " + " ".join(
                        map(shlex.quote, script.split(" "))
                    )
                    print("\n\n" + f"{' SCRIPT ':#^80}")
                    print(script)
                else:
                    print(f"{' SCRIPT ':#^80}")
                    print(script)
                runner = self.initialized_runner(script)
                # print all parameters
                print("\n\n" + f"{' PARAMETERS ':#^80}")
                print(runner.model.parameters)
                # print all expectation values
                print("\n\n" + f"{' NOMINAL EXPECTATION VALUES ':#^80}")
                try:
                    expectation_values = runner.model.nominal_expectation_values
                    max_key_length = max([len(k) for k in expectation_values.keys()])
                    for k, v in expectation_values.items():
                        print(f"{k:<{max_key_length}}   {v}")
                except NotImplementedError as msg:
                    warnings.warn(str(msg))
                return runner
            subprocess.call(shlex.split(script))


class NeymanConstructor(SubmitterLocal):
    """Neyman threshold constructor.

    Will not really submit a job.

    """

    allowed_special_args = ["free_name", "true_name", "confidence_levels"]

    def submit(
        self,
        free_name: str = "free",
        true_name: str = "true",
        confidence_levels: List[float] = [0.6827, 0.8, 0.9, 0.95],
    ):
        """Read the likelihood ratio from the output files and calculate the Neyman threshold. The
        threshold will be saved into a json file. The threshold will be sorted based on the elements
        of poi.

        Args:
            free_name: the name of the free hypothesis
            true_name: the name of the true hypothesis
            confidence_levels: the confidence levels to calculate the threshold
                0.6827 = stats.norm.cdf(1) - stats.norm.cdf(-1)

        Example:
            >>> data = json.load(open("limit_threshold.json")); print(json.dumps(data, indent=4))
            {
                "oyldh6bczx": {
                    "hashed_keys": {
                        "poi": "wimp_rate_multiplier",
                        "nominal_values": {
                            "wimp_mass": 0,
                            "livetime": 0.0
                        },
                        "generate_values": {},
                        "confidence_level": 0.9
                    },
                    "wimp_rate_multiplier": [
                        0.0,
                        1.0,
                        2.0
                    ],
                    "threshold": [
                        0.0,
                        0.0,
                        0.0
                    ],
                    "poi_expectation": [
                        null,
                        null,
                        null
                    ]
                },
            }

        """

        # overwrite the free_name, true_name and confidence_levels from the runner_args
        runner_args = next(self.merged_arguments_generator())
        if "free_name" in runner_args:
            free_name = runner_args["free_name"]
            self.logging.info(f"Overwrite free_name to {free_name}.")
        if "true_name" in runner_args:
            true_name = runner_args["true_name"]
            self.logging.info(f"Overwrite true_name to {true_name}.")
        if "confidence_levels" in runner_args:
            confidence_levels = runner_args["confidence_levels"]
            self.logging.info(f"Overwrite confidence_levels to {confidence_levels}.")

        # extract limit_threshold from the statistical_model_args
        limit_threshold = runner_args["statistical_model_args"].get("limit_threshold", None)
        if limit_threshold is None:
            raise ValueError("Please specify the limit_threshold at in_common.")
        if os.path.splitext(limit_threshold)[-1] != ".json":
            raise ValueError("The limit_threshold file should be a json file.")

        # calculate the threshold, iterate over the output files
        threshold = cast(Dict[str, Any], {})
        for runner_args in self.merged_arguments_generator():
            # check if the free_name, true_name and confidence_levels are consistent
            message = (
                " is not consistent in one of your runner arguments, "
                "please only specify it in in_common."
            )
            if runner_args.pop("free_name", free_name) != free_name:
                raise ValueError("free_name" + message)
            if runner_args.pop("true_name", true_name) != true_name:
                raise ValueError("true_name" + message)
            new_confidence_levels = runner_args.pop("confidence_levels", confidence_levels)
            mask = any([n_c != c for n_c, c in zip(new_confidence_levels, confidence_levels)])
            mask |= len(new_confidence_levels) != len(confidence_levels)
            if mask:
                raise ValueError("confidence_levels" + message)

            # prepare the needed nominal_values and generate_values
            nominal_values = deepcopy(runner_args["nominal_values"])
            generate_values = deepcopy(runner_args["generate_values"])
            needed_kwargs = {
                **nominal_values,
                **generate_values,
            }

            # get the output filename pattern
            output_filename_pattern = search_filename_pattern(
                runner_args["output_filename"].format(**needed_kwargs)
            )

            # read metadata including generate_values
            metadata = self._read_metadata(output_filename_pattern)

            # combine the generate_values from
            # the metadata(where the _rate_multiplier is already calculated)
            needed_kwargs = {**metadata["generate_values"], **needed_kwargs}

            # read poi and poi_expectation
            poi_value, poi_expectation = self._read_poi(metadata, **needed_kwargs)

            # read the likelihood ratio
            results = toyfiles_to_numpy(output_filename_pattern)
            llfree = results[free_name]["ll"]
            lltrue = results[true_name]["ll"]
            llrs = 2.0 * (llfree - lltrue)
            if llrs.min() < 0.0:
                mean_valid = (
                    results[free_name]["valid_fit"] & results[true_name]["valid_fit"]
                ).mean()
                self.logging.warning(
                    f"The lowest log likelihood ratio is negative {llrs.min():.02e}, "
                    f"total fraction of negative log likelihood ratio is "
                    f"{(llrs < 0.0).sum() / len(llrs):.02f}, "
                    f"total fraction of invalid fit is {1 - mean_valid:.02f}, "
                    f"the median of negative log likelihood ratios "
                    f"is {np.median(llrs[llrs < 0.0]):.02e}, "
                    f"there might be a problem in your fitting."
                )
            if len(llrs) < 1000:
                self.logging.warning(
                    "The number of toys is less than 1000, the threshold might not be accurate!"
                )

            # make sure no poi and poi_expectation in the hashed_keys
            generate_values.pop(self.poi, None)
            generate_values.pop("poi_expectation", None)
            nominal_values.pop(self.poi, None)
            nominal_values.pop("poi_expectation", None)

            # calculate the threshold given different confidence levels
            for confidence_level in confidence_levels:
                q_llr = np.percentile(llrs, 100.0 * confidence_level).item()
                if q_llr < 0.0:
                    self.logging.warning(
                        f"The threshold is negative ({q_llr}) for confidence_level "
                        f"{confidence_level}, there might be a problem in your fitting."
                    )
                hashed_keys = {
                    "poi": self.poi,
                    "nominal_values": nominal_values,
                    "generate_values": generate_values,
                    "confidence_level": confidence_level,
                }
                threshold_key = deterministic_hash(hashed_keys)
                if threshold_key not in threshold:
                    threshold_value = {
                        "hashed_keys": hashed_keys,
                        self.poi: [],
                        "threshold": [],
                        "poi_expectation": [],
                    }
                    threshold[threshold_key] = deepcopy(threshold_value)
                threshold[threshold_key][self.poi].append(poi_value)
                threshold[threshold_key]["threshold"].append(q_llr)
                threshold[threshold_key]["poi_expectation"].append(poi_expectation)

        # sort the threshold based on the elements of poi
        for k, v in threshold.items():
            paired = zip(v[self.poi], v["threshold"], v["poi_expectation"])
            # sort the pairs based on the elements of poi
            sorted_pairs = sorted(paired, key=lambda x: x[0])
            threshold[k][self.poi] = [x[0] for x in sorted_pairs]
            threshold[k]["threshold"] = [x[1] for x in sorted_pairs]
            threshold[k]["poi_expectation"] = [x[2] for x in sorted_pairs]

        # save the threshold into a json file
        with open(limit_threshold, mode="w") as f:
            json.dump(threshold, f, indent=4)
        print(f"Saving {limit_threshold}")

    def _read_metadata(self, output_filename_pattern):
        """Read metadata from the output files."""
        output_filename_list = sorted(glob(output_filename_pattern))
        metadata_list = get_metadata(output_filename_pattern)
        for m in metadata_list:
            m.pop("date", None)
        if len(set([deterministic_hash(m) for m in metadata_list])) != 1:
            raise ValueError(
                f"The metadata are not the same for all the {len(output_filename_list)} output!"
            )
        metadata = metadata_list[0]
        if metadata["poi"] != self.poi:
            raise ValueError(
                f"The poi in the metadata {metadata['poi']} is not "
                f"the same as the poi {self.poi}!"
            )
        return metadata

    def _read_poi(self, metadata, **kwargs):
        """Read poi and poi_expectation from the metadata, and check if the poi_expectation is
        consistent with the poi_expectation from the model."""
        poi_expectation = kwargs.pop("poi_expectation", None)
        poi_value = kwargs.get(self.poi, None)
        if poi_value is None:
            raise ValueError("Can not find the poi value in the generate_values in metadata!")
        # read expectation_values from metadata
        expectation_values = metadata["expectation_values"]
        # check if the poi_expectation is in expectation_values
        source = self.poi.replace("_rate_multiplier", "")
        if source not in expectation_values:
            warnings.warn(
                f"poi {self.poi} does not corresponds to any source in the model!"
                " so can not calculate the poi_expectation!"
            )
        else:
            _poi_expectation = expectation_values[source]
            if poi_expectation is not None:
                # check if the poi_expectation is consistent
                if not np.isclose(poi_expectation, _poi_expectation):
                    raise ValueError(
                        f"The poi_expectation from model {poi_expectation} is not "
                        f"the same as the poi_expectation from toymc {_poi_expectation}!"
                    )
            else:
                warnings.warn(
                    "Can not find the poi_expectation in the generate_values, "
                    "so will not check if the poi_expectation "
                    "are consistent!"
                )
                poi_expectation = _poi_expectation
        return poi_value, poi_expectation

    @staticmethod
    def build_interpolator(
        poi,
        threshold,
        generate_values,
        nominal_values,
        confidence_level,
    ):
        """Build interpolator from the limit_threshold file.

        Args:
            poi (str): parameter of interest
            generate_values (dict): generate values assumed to be in the limit_threshold file,
                but is actually hypothesis. It should not contain poi or poi_expectation.
            nominal_values (dict): nominal values of parameters
            confidence_level (float): confidence level
            threshold (dict): threshold read directly from limit_threshold file

        Raises:
            ValueError: if the limit_threshold file does not contain the thresholds for the
                generate_values

        """
        names = list(generate_values.keys())
        inputs = threshold.values()
        # filter out the inputs with different nominal_values
        inputs = [i for i in inputs if i["hashed_keys"]["nominal_values"] == nominal_values]
        # filter out the inputs with different confidence_level
        inputs = [i for i in inputs if i["hashed_keys"]["confidence_level"] == confidence_level]
        # filter out the inputs with different generate_values keys
        inputs = [
            i for i in inputs if set(i["hashed_keys"]["generate_values"].keys()) == set(names)
        ]

        if len(inputs) == 0:
            raise ValueError(
                f"limit_threshold file does not contain "
                f"any threshold for nominal_values {nominal_values} "
                f"confidence_level {confidence_level}, and generate_values keys {names}!"
            )

        # get poi list
        poi_values = inputs[0][poi]
        if any([set(i[poi]) != set(poi_values) for i in inputs]):
            raise ValueError(
                f"poi {poi} lists with nominal_values {nominal_values} and "
                f"confidence_level {confidence_level} "
                "are not all the same in the limit_threshold file!"
            )

        # in the following code, we assume the generate_values are the same in the limit_threshold
        # and names, points, bounds, within have same length, each element corresponds to one name

        # deduce dimension of interpolator and collect all threshold into values
        points = []
        for n in names:
            points.append(
                np.unique([i["hashed_keys"]["generate_values"][n] for i in inputs]).tolist()
            )
        # check if the poi_values have the same length in the limit_threshold file
        if any([len(i[poi]) != len(poi_values) for i in inputs]):
            raise ValueError(
                f"poi {poi} lists with nominal_values {nominal_values} and "
                f"confidence_level {confidence_level} "
                "do not the same length in the limit_threshold file!"
            )
        # check if limit_threshold file contains enough threshold
        size = reduce(operator.mul, [len(p) for p in points])
        if len(inputs) != size:
            raise ValueError(
                f"limit_threshold file does not contain "
                f"enough threshold for nominal_values {nominal_values} "
                f"confidence_level {confidence_level}, and generate_values keys {names}!"
            )
        # build the values array for interpolator
        values = np.empty((*[len(p) for p in points], len(poi_values)), dtype=float)
        for p in itertools.product(*points):
            indices = [pi.index(pii) for pii, pi in zip(p, points)]
            _threshold = [
                i["threshold"]
                for i in inputs
                if i["hashed_keys"]["generate_values"] == dict(zip(names, p))
            ]
            if len(_threshold) > 1:
                raise ValueError(
                    f"More than one threshold for nominal_values {nominal_values} "
                    f"confidence_level {confidence_level}, and generate_values keys {names}!"
                )
            values[tuple(indices)] = _threshold[0]

        interpolator = RegularGridInterpolator(
            points + [poi_values], values, method="linear", bounds_error=True
        )
        return interpolator

    @staticmethod
    def get_confidence_interval_thresholds(
        poi,
        hypotheses_values,
        limit_threshold,
        nominal_values,
        confidence_interval_kind,
        confidence_level,
        limit_threshold_interpolation,
        asymptotic_dof: Optional[int] = 1,
    ):
        """Get confidence interval threshold function from limit_threshold file. If the
        limit_threshold file does not contain the threshold, it will interpolate the threshold from
        the existing threshold, using the RegularGridInterpolator, so in this case the threshold is
        not exact.

        Args:
            poi (str): parameter of interest
            hypotheses_values (list): hypotheses values for statistical model,
                only contains Dict[str, float]
            limit_threshold (str): path to the limit_threshold file
            nominal_values (dict): nominal values of parameters
            confidence_level (float): confidence level
            limit_threshold_interpolation (bool): whether to interpolate the threshold from the
                existing threshold, if the limit_threshold file does not contain the threshold
            asymptotic_dof (int, optional (default=1)):
                degrees of freedom for asymptotic critical value

        """

        if limit_threshold is None:
            return [None] * len(hypotheses_values)

        threshold = load_json(limit_threshold)

        func_list: List[Optional[Callable]] = []
        for i_hypo in range(len(hypotheses_values)):
            hypothesis = hypotheses_values[i_hypo]

            # keys for hashing, should be in limit_threshold
            _nominal_values = deepcopy(nominal_values) if nominal_values else {}
            _generate_values = deepcopy(hypothesis) if hypothesis else {}
            hashed_keys = {
                "poi": poi,
                "nominal_values": deepcopy(_nominal_values),
                "generate_values": deepcopy(_generate_values),
                "confidence_level": confidence_level,
            }

            # make sure no poi and poi_expectation in the hashed_keys
            hashed_keys["generate_values"].pop(poi, None)
            hashed_keys["generate_values"].pop("poi_expectation", None)
            hashed_keys["nominal_values"].pop(poi, None)
            hashed_keys["nominal_values"].pop("poi_expectation", None)
            threshold_key = deterministic_hash(hashed_keys)

            if (threshold_key not in threshold) and (not limit_threshold_interpolation):
                raise ValueError(
                    f"Looking for hashed_keys {hashed_keys}, but limit_threshold file "
                    f"{limit_threshold} does not contain "
                    f"{threshold_key}! Please check the limit_threshold file or set "
                    f"limit_threshold_interpolation as true!"
                )

            # get the poi_values and threshold_values for interp1d
            if threshold_key in threshold:
                # if threshold_key in threshold, get the poi_values and threshold_values directly
                poi_values = threshold[threshold_key][poi]
                threshold_values = threshold[threshold_key]["threshold"]
            else:
                # if threshold_key not in threshold,
                # interpolate the threshold from the existing threshold
                if len(hashed_keys["generate_values"]) == 0:
                    warnings.warn(
                        f"If hypothesis {hypothesis} does not contain any values except "
                        f"poi, the nominal values should be in limit_threshold, so that "
                        f"the threshold can be interpolated from the existing threshold."
                    )
                    func_list.append(None)
                    continue

                names = list(hashed_keys["generate_values"].keys())
                interpolator = NeymanConstructor.build_interpolator(
                    poi,
                    threshold,
                    hashed_keys["generate_values"],
                    hashed_keys["nominal_values"],
                    confidence_level,
                )
                poi_values = interpolator.grid[-1]

                if set(hashed_keys["generate_values"].keys()) != set(names):
                    raise ValueError(
                        f"When using limit_threshold_interpolation, "
                        f"the keys of hypotheses should be "
                        f"the same for all hypotheses(expect str hypotheses)! "
                        f"But previously find {set(names)} and now find "
                        f"{set(hashed_keys['generate_values'].keys())}!"
                    )
                pts = [list(hashed_keys["generate_values"].values()) + [p] for p in poi_values]
                try:
                    threshold_values = interpolator(pts)
                except ValueError:
                    raise ValueError(
                        f"Interpolation failed for hypothesis. "
                        f"Maybe the limit_threshold file does not contain enough threshold, "
                        f"so that {hashed_keys['generate_values']} is out of bounds. "
                        f"Please check the limit_threshold file!"
                    )

            # if out of bounds, return the asymptotic critical value
            func = interp1d(
                poi_values,
                threshold_values,
                bounds_error=False,
                fill_value=asymptotic_critical_value(
                    confidence_interval_kind, confidence_level, asymptotic_dof
                ),
            )
            func_list.append(func)

        # check if the length of hypotheses and func_list are the same
        if len(hypotheses_values) != len(func_list):
            raise ValueError(
                "Something wrong with the length of hypotheses and func_list, "
                "please check the code!"
            )
        return func_list



import warnings
from typing import List, Dict, Callable, Optional, Union, cast
from pydoc import locate
import itertools
from copy import deepcopy

import numpy as np
import scipy.stats as stats
from blueice.likelihood import LogAncillaryLikelihood, LogLikelihoodSum
from inference_interface import dict_to_structured_array, structured_array_to_dict

from alea.model import StatisticalModel
from alea.parameters import Parameters
from alea.simulators import BlueiceDataGenerator
from alea.utils import ReadOnlyDict
from alea.utils import adapt_likelihood_config_for_blueice, get_template_folder_list, load_yaml


class BlueiceExtendedModel(StatisticalModel):
    """A statistical model based on blueice likelihoods.

    This class extends the ``StatisticalModel`` class and provides methods
    for generating data and computing likelihoods based on blueice.

    Attributes:
        parameters (Parameters): Parameters object containing the parameters of the model.
        data (dict): Data of the statistical model.
        is_data_set (bool): Whether data is set.
        _likelihood (LogLikelihoodSum): A blueice LogLikelihoodSum instance.
        likelihood_names (list): List of likelihood names.
        livetime_parameter_names (list): List of the name of the livetime of each term,
            None if not specified
        data_generators (list): List of data generators for each likelihood term.

    Args:
        parameter_definition (dict): A dictionary defining the model parameters.
        likelihood_config (dict): A dictionary defining the likelihood.

    Todo:
        analysis_space could be inferred from the data
        (assert that all sources have the same analysis_space)

    """

    def __init__(self, parameter_definition: dict, likelihood_config: dict, **kwargs):
        """Initializes the statistical model.

        Args:
            parameter_definition (dict): A dictionary defining the model parameters.
            likelihood_config (dict): A dictionary defining the likelihood.

        """
        super().__init__(parameter_definition=parameter_definition, **kwargs)
        self._likelihood = self._build_ll_from_config(
            likelihood_config, template_path=kwargs.get("template_path", None)
        )
        self.likelihood_names = [t["name"] for t in likelihood_config["likelihood_terms"]]
        self.likelihood_names.append("ancillary")
        self.livetime_parameter_names = [
            t.get("livetime_parameter", None) for t in likelihood_config["likelihood_terms"]
        ]
        self.livetime_parameter_names += [None]  # ancillary likelihood
        self.data_generators = self._build_data_generators()
        self._set_default_ptype()

    @classmethod
    def from_config(cls, config_file_path: str, **kwargs) -> "BlueiceExtendedModel":
        """Initializes the statistical model from a yaml config file.

        Args:
            config_file_path (str): Path to the yaml config file.

        Returns:
            BlueiceExtendedModel: Statistical model.

        """
        config = load_yaml(config_file_path)
        return cls(**{**config, **kwargs})

    @property
    def data(self) -> Union[dict, list]:
        """Return the data of the statistical model."""
        return super().data

    @data.setter
    def data(self, data: Union[dict, list]):
        """Overrides default setter. Will also set the data of the blueice ll. Data-sets are
        expected to be in the form of a list of one or more structured arrays representing the data-
        sets of one or more likelihood terms.

        Args:
            data (dict or list): Data of the statistical model.
                If data is a list, it must be a list of length len(self.likelihood_names) + 1.

        Raises:
            Warning: If data is not a list of length len(self.likelihood_names) + 1.

        Caution:
            The self._data is read-only, so you can not change the data after it is set.
            In order to change the data, you have to set the data again, using self.data = ***.

        """
        # iterate through all likelihood terms and set the science data in the blueice ll
        # last entry in data are the generate_values
        if isinstance(data, list):
            if len(data) != len(self.likelihood_names) + 1:
                if len(data) == len(self.likelihood_names):
                    warnings.warn(
                        f"If data is not a list of length {len(self.likelihood_names) + 1}, "
                        f"only the science data and ancillary will be set."
                    )
                else:
                    raise ValueError(
                        "You should at least provide data for all likelihood terms, "
                        "including science data and ancillary."
                    )
            data = dict(zip((self.likelihood_names + ["generate_values"])[: len(data)], data))
        for i, (dataset_name, d) in enumerate(data.items()):
            if dataset_name != "generate_values":
                ll_term = self.likelihood_list[i]
                if dataset_name != ll_term.pdf_base_config["name"]:
                    raise ValueError("Likelihood names do not match.")
                ll_term.set_data(d)

        self._data = ReadOnlyDict(data)
        self.is_data_set = True

    def get_source_name_list(self, likelihood_name: str) -> list:
        """Return a list of source names for a given likelihood term. The order is the same as used
        in the ``source`` column of the data, so this can be used to map the indices provided in the
        data to a source name.

        Args:
            likelihood_name (str): Name of the likelihood.

        Returns:
            list: List of source names.

        """
        ll_index = self.likelihood_names.index(likelihood_name)
        return self.likelihood_list[ll_index].source_name_list

    @property
    def all_source_names(self) -> list:
        """Return a set of possible source names from all likelihood terms.

        Args:
            likelihood_name (str): Name of the likelihood.
        Returns:
            set: set of source names.

        """
        source_names = set(
            itertools.chain.from_iterable([ll.source_name_list for ll in self.likelihood_list[:-1]])
        )
        return sorted(source_names)

    @property
    def likelihood_list(self) -> List:
        """Return a list of likelihood terms."""
        return self._likelihood.likelihood_list

    @property
    def likelihood_parameters(self) -> List:
        """Return a list of likelihood parameters."""
        return self._likelihood.likelihood_parameters

    def get_expectation_values(self, per_likelihood_term=False, **kwargs) -> dict:
        """Return total expectation values (summed over all likelihood terms with the same name)
        given a number of named parameters (kwargs)

        Args:
            per_likelihood_term (bool): If True, return expectation values
                per likelihood term. Otherwise, sum each source over all likelihood terms.
            kwargs: Named parameters

        Returns:
            dict: Dictionary of expectation values. If per_likelihood_term is True, the dictionary
                has the form {likelihood_name: {source_name: expectation_value, ...}, ...}.

        Todo:
            Make a self.likelihood_temrs dict with the likelihood names as keys and
            the corresponding likelihood terms as values.

        """
        ret = cast(Dict[str, Dict[str, float]], {})
        # prepare generate_values
        if not self.parameters.values_in_fit_limits(**kwargs):
            raise ValueError("Values are not within fit limits")
        generate_values = self.parameters(**kwargs)

        # ancillary likelihood does not contribute
        for ll_name, lt_name in zip(
            self.likelihood_names[:-1],
            self.livetime_parameter_names,
        ):
            ret[ll_name] = {}
            ll_index = self.likelihood_names.index(ll_name)
            lt = generate_values.pop(lt_name, None)
            # compute the mus
            self.data_generators[ll_index].compute_pdfs_and_mus(**generate_values, livetime_days=lt)
            mus = self.data_generators[ll_index].mus
            for n, mu in zip(self.likelihood_list[ll_index].source_name_list, mus):
                ret[ll_name][n] = mu
            # sort by source name
            ret[ll_name] = dict(sorted(ret[ll_name].items(), key=lambda item: item[0]))
        if not per_likelihood_term:
            # sum over sources with same names of all likelihood terms
            ret = {
                n: sum([ret[ll_name].get(n, 0.0) for ll_name in ret.keys()])  # type: ignore
                for n in self.all_source_names
            }

        return ret

    def get_source_histograms(self, likelihood_name: str, expected_events=False, **kwargs) -> dict:
        """Return the pdfs or histograms of all sources for a given likelihood term.

        Args:
            likelihood_name (str): Name of the likelihood term.
            expected_events (bool): If True, return the histograms containing
                the number of expected events.
            kwargs: Named parameters.

        Returns:
            dict: Dictionary containing a multihist object for each source.

        """
        if likelihood_name not in self.likelihood_names:
            raise ValueError(f"Likelihood {likelihood_name} not found.")
        elif likelihood_name == "ancillary":
            raise ValueError("No source histograms for ancillary likelihood.")

        ll_index = self.likelihood_names.index(likelihood_name)

        # prepare generate_values
        if not self.parameters.values_in_fit_limits(**kwargs):
            raise ValueError("Values are not within fit limits")
        generate_values = self.parameters(**kwargs)
        lt_name = self.livetime_parameter_names[ll_index]
        # change keyof lt_name to "livetime_days" if it is in the generate_values
        if lt_name in generate_values:
            generate_values["livetime_days"] = generate_values.pop(lt_name)

        # compute the pdfs
        self.data_generators[ll_index].compute_pdfs_and_mus(**generate_values)
        source_histograms = deepcopy(self.data_generators[ll_index].source_histograms)

        if expected_events:
            mus = self.data_generators[ll_index].mus
            for source_name, hist in source_histograms.items():
                source_index = self.get_source_name_list(likelihood_name).index(source_name)
                hist.histogram *= mus[source_index]
        # for unbinned likelihoods we need to divide by the bin volumes
        elif not expected_events and not self.data_generators[ll_index].binned:
            for hist in source_histograms.values():
                hist.histogram /= hist.bin_volumes()

        # sort the source_histograms by source name
        source_histograms = dict(sorted(source_histograms.items(), key=lambda item: item[0]))

        return source_histograms

    def _process_blueice_config(self, config, template_folder_list):
        """Process the blueice config from config."""
        pdf_base_config = adapt_likelihood_config_for_blueice(config, template_folder_list)
        pdf_base_config["livetime_days"] = self.parameters[
            pdf_base_config["livetime_parameter"]
        ].nominal_value
        for p in self.parameters:
            # adding the nominal rate values will screw things up in blueice!
            # So here we're just adding the nominal values of all other parameters
            if p.ptype != "rate":
                pdf_base_config[p.name] = pdf_base_config.get(p.name, p.nominal_value)

        # sanity checks
        for source in config["sources"]:
            if "name" not in source:
                raise ValueError("No name specified for source.")
            if "parameters" not in source:
                raise ValueError(f"No parameters specified for source {source['name']}.")
            if set(source.get("named_parameters", [])) - set(source["parameters"]):
                raise ValueError(
                    f"Named parameters {source['named_parameters']} are not all in the "
                    f"parameter list {source['parameters']} of source {source['name']}."
                )

        # add all parameters to extra_dont_hash for each source unless it is used:
        for i, source in enumerate(config["sources"]):
            parameters_to_ignore = self._get_parameters_to_ignore(source)
            # ignore all shape parameters known to this model not named specifically
            # in the source:
            pdf_base_config["sources"][i]["extra_dont_hash_settings"] = parameters_to_ignore

        # get blueice likelihood_config if it's given
        likelihood_config = config.get("likelihood_config", None)

        blueice_config = {
            "pdf_base_config": pdf_base_config,
            "likelihood_config": likelihood_config,
        }
        return blueice_config

    def _get_parameters_to_ignore(self, source):
        parameters_to_ignore: List[str] = [
            p.name
            for p in self.parameters
            if (p.ptype in ["shape", "index"]) and (p.name not in source["parameters"])
        ]
        # no efficiency affects PDF:
        parameters_to_ignore += [p.name for p in self.parameters if (p.ptype == "efficiency")]
        parameters_to_ignore += source.get("extra_dont_hash_settings", [])
        return parameters_to_ignore

    def _build_ll_from_config(
        self, likelihood_config: dict, template_path: Optional[str] = None
    ) -> "LogLikelihoodSum":
        """Iterate through all likelihood terms and build blueice likelihood instances.

        Args:
            likelihood_config (dict): A dictionary defining the likelihood.

        Returns:
            LogLikelihoodSum: A blueice LogLikelihoodSum instance.

        """
        lls = []

        template_folder_list = get_template_folder_list(
            likelihood_config, extra_template_path=template_path
        )

        # Iterate through each likelihood term in the configuration
        for config in likelihood_config["likelihood_terms"]:
            blueice_config = self._process_blueice_config(config, template_folder_list)

            likelihood_class = cast(Callable, locate(config["likelihood_type"]))
            if likelihood_class is None:
                raise ValueError(f"Could not find {config['likelihood_type']}!")
            ll = likelihood_class(**blueice_config)

            for source in config["sources"]:
                # set rate parameters
                rate_parameters = [
                    p for p in source["parameters"] if self.parameters[p].ptype == "rate"
                ]
                if len(rate_parameters) != 1:
                    raise ValueError(
                        f"Source {source['name']} must have exactly one rate parameter."
                    )
                rate_parameter = rate_parameters[0]
                if rate_parameter.endswith("_rate_multiplier"):
                    rate_parameter = rate_parameter.replace("_rate_multiplier", "")
                    # The ancillary term is handled in CustomAncillaryLikelihood
                    ll.add_rate_parameter(rate_parameter, log_prior=None)
                else:
                    raise NotImplementedError(
                        "Only rate multipliers that end on _rate_multiplier"
                        " are currently supported."
                    )

                # set efficiency parameters
                if source.get("efficiency_name", None):
                    self._set_efficiency(source, ll)

                # set shape parameters
                shape_parameters = [
                    p
                    for p in source["parameters"]
                    if self.parameters[p].ptype in ["shape", "index"]
                ]
                for p in shape_parameters:
                    anchors = self.parameters[p].blueice_anchors
                    if anchors is None:
                        raise ValueError(f"Shape parameter {p} does not have any anchors.")
                    # The ancillary term is handled in CustomAncillaryLikelihood
                    ll.add_shape_parameter(p, anchors=anchors, log_prior=None)

            ll.prepare()
            lls.append(ll)

        # ancillary likelihood
        ll = CustomAncillaryLikelihood(self.parameters.with_uncertainty)
        lls.append(ll)

        likelihood_weights = likelihood_config.get("likelihood_weights", None)
        return LogLikelihoodSum(lls, likelihood_weights=likelihood_weights)

    def _build_data_generators(self) -> list:
        """Build data generators for all likelihood terms.

        Returns:
            list: List of data generators for each likelihood term.

        Todo:
            Also implement data generator for ancillary ll term.

        """
        # last one is AncillaryLikelihood
        data_generators = []
        for ll_term in self.likelihood_list[:-1]:
            methods = [s.config["pdf_interpolation_method"] for s in ll_term.base_model.sources]
            # make sure that all sources have the same pdf_interpolation_method
            if len(set(methods)) != 1:
                raise ValueError("All sources must have the same pdf_interpolation_method.")
            method = methods[0]
            if method == "piecewise":
                data_generators.append(BlueiceDataGenerator(ll_term))
            elif method == "linear":
                raise NotImplementedError(
                    "Linear interpolation is not yet supported."
                    " Choose piecewise as pdf_interpolation_method."
                )
            else:
                raise ValueError(f"Unknown pdf_interpolation_method {method}.")
        return data_generators

    def _ll(self, **generate_values) -> float:
        livetime_days = [generate_values.get(ln, None) for ln in self.livetime_parameter_names]
        return self._likelihood(livetime_days=livetime_days, **generate_values)

    def _generate_data(self, **generate_values) -> dict:
        """Generate data for all likelihood terms and ancillary likelihood.

        Keyword Args:
            generate_values (dict): A dictionary of parameter values.

        Returns:
            dict: A dict of data-sets,
            with key of the likelihood term name, "ancillary" and "generate_values".

        """
        # generate_values are already filtered and filled by the nominal values
        data = self._generate_science_data(**generate_values)
        ancillary_keys = self.parameters.with_uncertainty.names
        generate_values_anc = {k: v for k, v in generate_values.items() if k in ancillary_keys}
        data["ancillary"] = self._generate_ancillary(**generate_values_anc)
        data["generate_values"] = dict_to_structured_array(generate_values)
        return data

    def store_data(self, file_name, data_list, data_name_list=None, metadata=None):
        """Store data in a file.

        Append the generate_values to the data_name_list.

        """
        if data_name_list is None:
            data_name_list = self.likelihood_names + ["generate_values"]
        super().store_data(file_name, data_list, data_name_list, metadata)

    def _generate_science_data(self, **generate_values) -> dict:
        """Generate the science data for all likelihood terms except the ancillary likelihood."""
        livetime_days = [generate_values.get(ln, None) for ln in self.livetime_parameter_names]
        science_data = [
            gen.simulate(livetime_days=lt, **generate_values)
            for gen, lt in zip(self.data_generators, livetime_days)
        ]
        return dict(zip(self.likelihood_names[:-1], science_data))

    def _generate_ancillary(self, **generate_values) -> dict:
        """Generate data for the ancillary likelihood.

        Keyword Args:
            generate_values (dict): A dictionary of parameter values.

        Returns:
            numpy.array: A numpy structured array of ancillary measurements.

        """
        ancillary = {}
        anc_ll = self.likelihood_list[-1]
        ancillary_generators = anc_ll._get_constraint_functions(**generate_values)
        for name, gen in ancillary_generators.items():
            parameter_meas = gen.rvs()
            # correct parameter_meas if out of bounds
            param = self.parameters[name]
            if not param.value_in_fit_limits(parameter_meas):
                if param.fit_limits[0] is not None and parameter_meas < param.fit_limits[0]:
                    parameter_meas = param.fit_limits[0]
                elif param.fit_limits[1] is not None and parameter_meas > param.fit_limits[1]:
                    parameter_meas = param.fit_limits[1]
            ancillary[name] = parameter_meas

        return dict_to_structured_array(ancillary)

    def _set_efficiency(self, source: dict, ll):
        """Set the efficiency of a source in the blueice ll.

        Args:
            source (dict): A dictionary defining the source.
            ll (LogLikelihood): A blueice LogLikelihood instance.

        Raises:
            ValueError: If the efficiency_name is not specified in the source.

        """
        if "efficiency_name" not in source:
            raise ValueError(f"Unspecified efficiency_name for source {source['name']:s}")
        efficiency_name = source["efficiency_name"]

        if efficiency_name not in source["parameters"]:
            raise ValueError(
                f"The efficiency_name for source {source['name']:s} is not in its parameter list"
            )
        efficiency_parameter = self.parameters[efficiency_name]

        if efficiency_parameter.ptype != "efficiency":
            raise ValueError(f"The parameter {efficiency_name:s} must be an efficiency")
        limits = efficiency_parameter.fit_limits

        if limits[0] < 0:
            raise ValueError(
                f"Efficiency parameters including {efficiency_name:s}"
                " must be constrained to be nonnegative"
            )
        if ~np.isfinite(limits[1]):
            raise ValueError(
                f"Efficiency parameters including {efficiency_name:s}"
                " must be constrained to be finite"
            )
        ll.add_shape_parameter(efficiency_name, anchors=(limits[0], limits[1]))

    def _set_default_ptype(self):
        """Check if all parameters have a ptype that is in the list of allowed ptypes.

        If no ptype is specified, set the default ptype "needs_reinit".

        """
        allowed_ptypes = ["rate", "shape", "index", "efficiency", "livetime", "needs_reinit"]
        default_ptype = "needs_reinit"
        for p in self.parameters:
            if p.ptype is None:
                p.ptype = default_ptype
            elif p.ptype not in allowed_ptypes:
                raise ValueError(
                    f"Parameter {p.name} has ptype {p.ptype} which is not in the list of "
                    f"allowed ptypes: {allowed_ptypes}."
                )

    def store_real_data(self, file_name: str, real_data_list: list, metadata=None):
        """Store real data in a file with toydata format.

        Args:
            file_name (str): Name of the file.
            real_data_list (list): List of np.array of real data.

        """
        # check if real_data_list has the correct length
        if len(real_data_list) != len(self.likelihood_names) - 1:
            raise ValueError(
                f"real_data_list must have length {len(self.likelihood_names) - 1} "
                f"according to the number of likelihood terms in the model, "
                f"but has length {len(real_data_list)}."
            )
        # check if the dtypes of the real data match the dtypes of the data generators
        expected_dtypes = [np.dtype(gen.dtype) for gen in self.data_generators]
        if any([r_d.dtype != e_d for r_d, e_d in zip(real_data_list, expected_dtypes)]):
            raise ValueError(
                "The dtypes of the real data do not match the dtypes of the data generators."
            )
        # set ancillary_measurements to nominal values
        _ancillary = self.parameters.with_uncertainty.nominal_values
        if None in _ancillary.values():
            raise ValueError(
                "The nominal values of the ancillary measurements are not set. "
                "Please provide nominal values for all ancillary measurements."
            )
        ancillary = dict_to_structured_array(_ancillary)
        # combine all data
        data_name_list = self.likelihood_names
        data_list = real_data_list + [ancillary]
        real_data = [dict(zip(data_name_list, data_list))]
        self.store_data(file_name, real_data, self.likelihood_names, metadata=metadata)


class CustomAncillaryLikelihood(LogAncillaryLikelihood):
    """Custom ancillary likelihood that can be used to add constraint terms for parameters of the
    likelihood.

    Attributes:
        parameters (Parameters): Parameters object containing the parameters to be constrained.
        constraint_functions (dict): Dict of constraint functions for all ancillary parameters.

    """

    def __init__(self, parameters: Parameters):
        """Initialize the CustomAncillaryLikelihood."""
        self.parameters = parameters
        # check that there are no None values in the uncertainties dict
        if set(self.parameters.uncertainties.keys()) != set(self.parameters.names):
            raise ValueError("The uncertainties dict must contain all parameters as keys.")
        parameter_list = self.parameters.names

        self.constraint_functions = self._get_constraint_functions()
        super().__init__(
            func=self.ancillary_sum,
            parameter_list=parameter_list,
            config=self.parameters.nominal_values,
        )
        self.pdf_base_config["name"] = "ancillary"

    @property
    def constraint_terms(self) -> dict:
        """Dict of all constraint terms (logpdf of constraint functions) of the ancillary
        likelihood.

        Returns:
            dict: Dict of all constraint terms function.

        """
        return {name: func.logpdf for name, func in self.constraint_functions.items()}

    def set_data(self, d: np.array):
        """Set the data of the ancillary likelihood (ancillary measurements).

        Args:
            d (numpy.array): Data of ancillary measurements, stored as numpy array.

        """
        # This results in shifted constraint terms.
        d_dict = structured_array_to_dict(d)
        if set(d_dict.keys()) != set(self.parameters.names):
            raise ValueError(
                "The data dict must contain all parameters as keys in CustomAncillaryLikelihood. "
                f"But {set(d_dict.keys())} is provided and "
                f"{set(self.parameters.names)} is expected."
            )
        self.constraint_functions = self._get_constraint_functions(**d_dict)

    def ancillary_sum(self, evaluate_at: dict) -> float:
        """Return the sum of all constraint terms.

        Args:
            evaluate_at (dict): Values of the ancillary measurements.

        Returns:
            float: Sum of all constraint terms.

        """
        evaluated_constraint_terms = [
            term(evaluate_at[name]) for name, term in self.constraint_terms.items()
        ]
        return np.sum(evaluated_constraint_terms)

    def _get_constraint_functions(self, **generate_values) -> dict:
        """Get callable constraint functions for all ancillary parameters.

        Keyword Args:
            generate_values (dict): A dictionary of parameter values.

        Returns:
            dict: Dict of constraint functions for all ancillary parameters.

        Todo:
            Implement str-type uncertainties.

        """
        central_values = self.parameters(**generate_values)
        constraint_functions = {}
        for name, uncertainty in self.parameters.uncertainties.items():
            if isinstance(uncertainty, (float, int)):
                param = self.parameters[name]
                if param.relative_uncertainty:
                    if param.nominal_value is None:
                        raise ValueError(
                            f"Relative uncertainty of parameter {name} is set to {uncertainty} "
                            "but nominal value is None. "
                            "Please provide a nominal value."
                        )
                    if param.nominal_value == 0:
                        warnings.warn(
                            f"Relative uncertainty of parameter {name} is set to {uncertainty} "
                            "but nominal value is 0. "
                            "This will result in a relative uncertainty of 0."
                        )
                    uncertainty *= param.nominal_value
                func = stats.norm(central_values[name], uncertainty)
            elif hasattr(uncertainty, "logpdf") and hasattr(uncertainty, "rvs"):
                warnings.warn(
                    f"Uncertainty of {name} is a string-based uncertainty. "
                    "It is frozen and its argument(s) cannot be changed as ancillary measurement."
                )
                func = uncertainty
            else:
                raise NotImplementedError(
                    f"Uncertainty {uncertainty} is not understandable. "
                    "Only float, int, and scipy.stats distributions are supported."
                )
            constraint_functions[name] = func
        return constraint_functions



import setuptools


def open_requirements(path):
    with open(path) as f:
        requires = [
            r.split("/")[-1] if r.startswith("git+") else r
            for r in f.read().splitlines()
            if not r.startswith("-")
        ]
    return requires


requires = open_requirements("requirements.txt")

with open("README.md") as file:
    readme = file.read()

with open("HISTORY.md") as file:
    history = file.read()


setuptools.setup(
    name="alea-inference",
    version="0.2.8",
    description="A tool to perform toyMC-based inference constructions",
    author="Alea contributors, the XENON collaboration",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    setup_requires=["pytest-runner"],
    install_requires=requires,
    python_requires=">=3.8",
    packages=setuptools.find_packages(),
    package_data={
        "alea": [
            "examples/configs/*",
            "examples/templates/*",
        ],
    },
    scripts=[
        "bin/alea-run_toymc",
        "bin/alea-submission",
    ],
    url="https://github.com/XENONnT/alea",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    zip_safe=False,
)



import time
import inspect
from copy import deepcopy
from typing import Optional, Dict, Union
from datetime import datetime
import warnings

from tqdm import tqdm
import numpy as np
from inference_interface import toydata_from_file, numpy_to_toyfile

from alea.model import StatisticalModel
from alea.utils import load_yaml


class Runner:
    """Runner manipulates statistical model and toydata.

        - initialize the statistical model
        - generate or reads toy data
        - save toy data if needed
        - fit fittable parameters
        - write the output file
    One toyfile can contain multiple toydata, but all of them are from the same generate_values.

    Attributes:
        model (StatisticalModel): statistical model instance
        poi (str): parameter of interest
        hypotheses (list): list of hypotheses
        common_hypothesis (dict): common hypothesis, the values are copied to each hypothesis
        generate_values (dict): generate values for toydata
        nominal_values (dict): nominal values of parameters
        _compute_confidence_interval (bool): whether compute confidence interval
        _n_mc (int): number of Monte Carlo
        _toydata_filename (str): toydata filename
        _toydata_mode (str): toydata mode, 'read', 'generate', 'generate_and_store', 'no_toydata'
        _metadata (dict): metadata, if None, it is set to {}
        _output_filename (str): output filename
        _result_names (list): list of result names
        _result_dtype (list): list of result dtypes
        _hypotheses_values (list): list of values for hypotheses

    Args:
        statistical_model (str): statistical model class name
        poi (str): parameter of interest
        hypotheses (list): list of hypotheses
        n_mc (int): number of Monte Carlo
        common_hypothesis (dict, optional (default=None)):
            common hypothesis, the values are copied to each hypothesis
        generate_values (Dict[str, float], optional (default=None)):
            generate values of toydata. If None, toydata depend on statistical model.
        nominal_values (dict, optional (default=None)):
            nominal values of parameters. If None, nothing will be assigned to model.
        statistical_model_config (str, optional (default=None)):
            statistical model configuration filename
        parameter_definition (dict or list, optional (default=None)): parameter definition
        statistical_model_args (dict, optional (default={})): arguments for statistical model
        likelihood_config (dict, optional (default=None)): likelihood configuration
        compute_confidence_interval (bool, optional (default=False)):
            whether compute confidence interval
        confidence_level (float, optional (default=0.9)): confidence level
        confidence_interval_kind (str, optional (default='central')):
            kind of confidence interval, choice from 'central', 'upper' or 'lower'
        fit_strategy (dict, optional (default=None)): fit strategy dictionary.
            If None, the default fit strategy of the model will be used.
        toydata_mode (str, optional (default='generate_and_store')):
            toydata mode, choice from 'read', 'generate', 'generate_and_store', 'no_toydata'
        toydata_filename (str, optional (default=None)): toydata filename
        only_toydata (bool, optional (default=False)): whether only generate toydata
        output_filename (str, optional (default='test_toymc.ii.h5')): output filename
        seed (int, optional (default=None)): random seed for runners before generating toydata
        metadata (dict, optional (default=None)): metadata to be saved in output file

    """

    def __init__(
        self,
        statistical_model: str = "alea.examples.gaussian_model.GaussianModel",
        poi: str = "mu",
        hypotheses: list = ["free"],
        n_mc: int = 1,
        common_hypothesis: Optional[dict] = None,
        generate_values: Optional[Dict[str, float]] = None,
        nominal_values: Optional[dict] = None,
        statistical_model_config: Optional[str] = None,
        parameter_definition: Optional[Union[dict, list]] = None,
        statistical_model_args: Optional[dict] = None,
        likelihood_config: Optional[dict] = None,
        compute_confidence_interval: bool = False,
        confidence_level: float = 0.9,
        confidence_interval_kind: str = "central",
        fit_strategy: Optional[dict] = None,
        toydata_mode: str = "generate_and_store",
        toydata_filename: str = "test_toydata_filename.ii.h5",
        only_toydata: bool = False,
        output_filename: str = "test_output_filename.ii.h5",
        seed: Optional[int] = None,
        metadata: Optional[dict] = None,
    ):
        """Initialize statistical model, parameters list, and generate values list."""
        self.poi = poi

        statistical_model_class = StatisticalModel.get_model_from_name(statistical_model)

        # if statistical_model_config is provided
        # overwrite parameter_definition and likelihood_config
        if statistical_model_config is not None:
            model_config = load_yaml(statistical_model_config)
            if parameter_definition is not None:
                raise ValueError(
                    "parameter_definition is duplicated, "
                    "because statistical_model_config is provided!"
                )
            if likelihood_config is not None:
                raise ValueError(
                    "likelihood_config is duplicated, "
                    "because statistical_model_config is provided!"
                )
            parameter_definition = model_config["parameter_definition"]
            likelihood_config = model_config["likelihood_config"]
            # in case of fit_strategy is provided in both
            # statistical_model_config and arguments fit_strategy in
            # arguments will overwrite fit_strategy in statistical_model_config
            fit_strategy = {**model_config.get("fit_strategy", {}), **(fit_strategy or {})}

        # update nominal_values into statistical_model_args
        if statistical_model_args is None:
            statistical_model_args = {}
        # nominal_values is keyword argument
        self.nominal_values = nominal_values if nominal_values else {}
        # initialize nominal_values only once
        statistical_model_args["nominal_values"] = self.nominal_values
        # likelihood_config is keyword argument, because not all statistical model needs it
        statistical_model_args["likelihood_config"] = likelihood_config

        # initialize statistical model
        self.model = statistical_model_class(
            parameter_definition=parameter_definition,
            confidence_level=confidence_level,
            confidence_interval_kind=confidence_interval_kind,
            fit_strategy=fit_strategy,
            **statistical_model_args,
        )

        self.hypotheses = hypotheses if hypotheses else []
        self.common_hypothesis = common_hypothesis if common_hypothesis else {}
        self.generate_values = generate_values if generate_values else {}
        self._compute_confidence_interval = compute_confidence_interval
        self._n_mc = n_mc
        self._toydata_filename = toydata_filename
        self._toydata_mode = toydata_mode
        self._output_filename = output_filename
        self.only_toydata = only_toydata
        self.seed = seed
        self._metadata = metadata if metadata else {}

        self._result_names, self._result_dtype = self._get_parameter_list()

        self._hypotheses_values = self._get_hypotheses()

        # find confidence_interval_thresholds function for the hypotheses
        from alea.submitters.local import NeymanConstructor

        self.confidence_interval_thresholds = NeymanConstructor.get_confidence_interval_thresholds(
            self.poi,
            self._hypotheses_values,
            statistical_model_args.get("limit_threshold", None),
            nominal_values,
            confidence_interval_kind,
            confidence_level,
            statistical_model_args.get("limit_threshold_interpolation", False),
            statistical_model_args.get("asymptotic_dof", 1),
        )

    def pre_process_poi(self, value, attribute_name):
        """Pre-process of poi_expectation for some attributes of runner."""
        if not all([isinstance(v, (float, int)) for v in value.values()]):
            raise ValueError(
                f"{attribute_name} should be a dict of float! But {value} is provided."
            )
        # update poi according to poi_expectation
        if "poi_expectation" in value:
            value = self.update_poi(self.model, self.poi, value, self.nominal_values)
        return value

    @property
    def generate_values(self) -> Dict[str, float]:
        return self._generate_values

    @generate_values.setter
    def generate_values(self, value: Dict[str, float]) -> None:
        if "poi_expectation" in value:
            self.input_poi_expectation = True
        else:
            self.input_poi_expectation = False
        # update poi according to poi_expectation
        self._generate_values = self.pre_process_poi(value, "generate_values")

    @property
    def common_hypothesis(self) -> Dict[str, float]:
        return self._common_hypothesis

    @common_hypothesis.setter
    def common_hypothesis(self, value: Dict[str, float]) -> None:
        # update poi according to poi_expectation
        self._common_hypothesis = self.pre_process_poi(value, "common_hypothesis")

    @property
    def hypotheses(self) -> list:
        return self._hypotheses

    @hypotheses.setter
    def hypotheses(self, values: list) -> None:
        # update poi according to poi_expectation
        for i in range(len(values)):
            if isinstance(values[i], dict):
                values[i] = self.pre_process_poi(values[i], "hypothesis")
        self._hypotheses = values

    @staticmethod
    def runner_arguments():
        """Get runner arguments and annotations."""
        # find run toyMC default args and annotations:
        # reference: https://docs.python.org/3/library/inspect.html#inspect.getfullargspec
        (
            args,
            varargs,
            varkw,
            defaults,
            kwonlyargs,
            kwonlydefaults,
            annotations,
        ) = inspect.getfullargspec(Runner.__init__)
        # skip the first one because it is self(Runner itself)
        default_args = dict(zip(args[1:], defaults))
        return args, default_args, annotations

    @staticmethod
    def update_poi(
        model, poi: str, generate_values: Dict[str, float], nominal_values: Dict[str, float] = {}
    ):
        """Update the poi according to poi_expectation. First, it will check if poi_expectation is
        provided, if not so, it will do nothing. Second, it will check if poi is provided, if so, it
        will raise error. Third, it will check if poi ends with _rate_multiplier, if not so, it will
        raise error. Finally, it will update poi to the correct value according to poi_expectation
        using the get_expectation_values method of model, under specified nominal_values.

        Args:
            poi (str): parameter of interest
            generate_values (dict): generate values of toydata,
                it can contain "poi_expectation"
            nominal_values (dict): nominal values of parameters

        Caution:
            The expectation is evaluated under nominal_values in each batch.

        """
        if poi in generate_values:
            raise ValueError(
                f"You can not specify both {poi} "
                f"along with poi_expectation, "
                f"because {poi} will be updated according to poi_expectation."
            )
        if not poi.endswith("_rate_multiplier"):
            raise ValueError(
                f"poi {poi} should end with _rate_multiplier, "
                "if poi_expectation is provided, because you want to update "
                "the generate_values according to the expectations."
            )
        generate_values_copy = deepcopy(generate_values)
        generate_values_copy.pop("poi_expectation")
        expectation_values = model.get_expectation_values(
            **{**generate_values_copy, **nominal_values}
        )
        component = poi.replace("_rate_multiplier", "")
        nominal_expectation = expectation_values[component]
        poi_expectation = generate_values["poi_expectation"]
        ratio = poi_expectation / nominal_expectation
        # update poi to the correct value
        generate_values.pop("poi_expectation")
        generate_values[poi] = ratio
        return generate_values

    def _get_parameter_list(self):
        """Get parameter list and result list from statistical model."""
        parameter_list = sorted(self.model.get_parameter_list())
        # add likelihood, lower limit, upper limit, and the migrad valid fit bool
        result_names = parameter_list + ["ll", "dl", "ul"]
        result_dtype = [(n, float) for n in result_names]
        result_names += ["valid_fit"]
        result_dtype += [("valid_fit", bool)]
        return result_names, result_dtype

    def _get_hypotheses(self):
        """Get generate values list from hypotheses.

        Caution:
            When free hypothesis is provided, it should be the first hypothesis.
            Free hypothesis means that all parameters are free to fit, it will
            not use common_hypothesis!

        """
        allowed_hypothesis_strs = ["zero", "true", "free"]
        hypotheses_values = []
        hypotheses = deepcopy(self.hypotheses)
        if len(hypotheses) == 0:
            raise ValueError("hypotheses should not be empty!")
        if "free" not in hypotheses and self._compute_confidence_interval:
            warnings.warn(
                f"If free hypothesis is not provided for confidence interval calculation, "
                f"the first hypothesis {hypotheses[0]} will be used for confidence "
                f"interval calculation."
            )
        if "free" in hypotheses and hypotheses.index("free") != 0:
            raise ValueError("free hypothesis should be the first hypothesis!")

        for hypothesis in hypotheses:
            # translate hypothesis
            if hypothesis == "free":
                # if free hypothesis, will not use common_hypothesis
                h = {}
            elif hypothesis == "zero":
                # there is no poi
                h = deepcopy(self.common_hypothesis)
                h.update({self.poi: 0.0})
            elif hypothesis == "true":
                # the true poi is used
                if self.poi not in self.generate_values:
                    raise ValueError(
                        f"{self.poi} should be provided in generate_values",
                    )
                h = deepcopy(self.common_hypothesis)
                h.update(
                    {
                        self.poi: self.generate_values.get(self.poi),
                    }
                )
            else:
                if not isinstance(hypothesis, dict):
                    raise ValueError(
                        f"If str hypothesis is not {allowed_hypothesis_strs}, it should be a dict!"
                    )
                h = deepcopy(self.common_hypothesis)
                h.update(hypothesis)

            if not all([isinstance(v, (float, int)) for v in h.values()]):
                raise ValueError(f"hypothesis should be a dict of float! But {h} is provided.")
            hypotheses_values.append(h)

        # check if the length of hypotheses and hypotheses_values are the same
        if len(hypotheses) != len(hypotheses_values):
            raise ValueError(
                "Something wrong with the length of hypotheses and hypotheses_values, "
                "please check the code!"
            )
        return hypotheses_values

    def write_output(self, results):
        """Write output file with metadata."""
        metadata = deepcopy(self._metadata)

        n_hypo = len(self._hypotheses_values)
        result_names = [f"{i:0{len(str(n_hypo))}d}" for i in range(n_hypo)]
        for i, ea in enumerate(self.hypotheses):
            if isinstance(ea, str) and (ea in {"free", "zero", "true"}):
                result_names[i] = ea

        metadata["date"] = datetime.now().strftime("%Y%m%d_%H:%M:%S")
        metadata["poi"] = self.poi
        metadata["common_hypothesis"] = self.common_hypothesis
        metadata["generate_values"] = self.generate_values
        metadata["nominal_values"] = self.nominal_values
        metadata["seed"] = self.seed
        try:
            metadata["expectation_values"] = self.model.get_expectation_values(
                **self.generate_values
            )
        except NotImplementedError:
            metadata["expectation_values"] = {}

        array_metadatas = [{"hypotheses_values": ea} for ea in self._hypotheses_values]
        numpy_arrays_and_names = [(r, rn) for r, rn in zip(results, result_names)]

        print(f"Saving {self._output_filename}")
        numpy_to_toyfile(
            self._output_filename,
            numpy_arrays_and_names=numpy_arrays_and_names,
            metadata=metadata,
            array_metadatas=array_metadatas,
        )

    def read_toydata(self):
        """Read toydata from file."""
        toydata, toydata_names = toydata_from_file(self._toydata_filename)
        return toydata, toydata_names

    def store_toydata(self, toydata, toydata_names):
        """Write toydata to file.

        If toydata is a list of dict, convert it to a list of list.

        """
        print(f"Saving {self._toydata_filename}")
        self.model.store_data(self._toydata_filename, toydata, toydata_names)

    def data_generator(self):
        """Generate, save or read toydata."""
        # set seed
        np.random.seed(self.seed)
        # check toydata mode
        if self._toydata_mode not in {
            "read",
            "generate",
            "generate_and_store",
            "no_toydata",
        }:
            raise ValueError(f"Unknown toydata mode: {self._toydata_mode}")
        if self.only_toydata and self._toydata_mode != "generate_and_store":
            raise ValueError(
                f"only_toydata is True, you should only generate_and_store, "
                f"but toydata_mode is {self._toydata_mode}!"
            )
        # check toydata file size
        if self._toydata_mode == "read":
            toydata, toydata_names = self.read_toydata()
            if len(toydata) < self._n_mc:
                raise ValueError(
                    f"Number of stored toydata {len(toydata)} is "
                    f"less than number of Monte Carlo {self._n_mc}!"
                )
            elif len(toydata) > self._n_mc:
                warnings.warn(
                    f"Number of stored toydata {len(toydata)} is "
                    f"larger than number of Monte Carlo {self._n_mc}."
                )
        else:
            toydata = []
            toydata_names = None
        # generate toydata
        for i_mc in range(self._n_mc):
            if self._toydata_mode == "generate" or self._toydata_mode == "generate_and_store":
                # generate toydata
                data = self.model.generate_data(**self.generate_values)
                # set fit guesses as generate values
                self.model.set_fit_guesses(**self.generate_values)
                if self._toydata_mode == "generate_and_store":
                    # append toydata
                    toydata.append(data)
            elif self._toydata_mode == "read":
                data = toydata[i_mc]
            elif self._toydata_mode == "no_toydata":
                data = None
            yield data
        # save toydata
        if self._toydata_mode == "generate_and_store":
            self.store_toydata(toydata, toydata_names)

    def simulate(self):
        """Only generate toydata."""
        all(tqdm(self.data_generator(), total=self._n_mc))

    def simulate_and_fit(self):
        """
        For each Monte Carlo:
            - run toy simulation a specified toydata mode and generate values.
            - loop over hypotheses.

        Todo:
            Implement per-hypothesis switching on whether to compute confidence intervals
        """
        results = [np.zeros(self._n_mc, dtype=self._result_dtype) for _ in self._hypotheses_values]
        for i_mc, data in tqdm(enumerate(self.data_generator()), total=self._n_mc):
            self.model.data = data
            fit_results = []
            for i_hypo, hypothesis_values in enumerate(self._hypotheses_values):
                fit_result, max_llh = self.model.fit(**hypothesis_values)
                fit_result["ll"] = max_llh
                fit_result["valid_fit"] = self.model.minuit_object.valid

                if self._compute_confidence_interval and (self.poi not in hypothesis_values):
                    # hypothesis_values should only be a fittable subset of parameters
                    non_fittable = set(hypothesis_values.keys()) - set(
                        self.model.parameters.fittable
                    )
                    if non_fittable:
                        raise ValueError(
                            f"The hypothesis {hypothesis_values} "
                            f"should only be a subset of the fittable parameters "
                            f"{self.model.parameters.fittable} in the statistical model. "
                            f"Because the non-fittable parameters {non_fittable} are fixed "
                            f"to nominal values in 'free' hypothesis."
                        )
                    dl, ul = self.model.confidence_interval(
                        poi_name=self.poi,
                        best_fit_args=self._hypotheses_values[0],
                        confidence_interval_args=hypothesis_values,
                        confidence_interval_threshold=self.confidence_interval_thresholds[i_hypo],
                    )
                else:
                    dl, ul = np.nan, np.nan
                fit_result["dl"] = dl
                fit_result["ul"] = ul

                fit_results.append(fit_result)
            # assign fitting results
            for fit_result, result_array in zip(fit_results, results):
                result_array[i_mc] = tuple(fit_result[pn] for pn in self._result_names)
        return results

    def run(self):
        """Run toy simulation.

        If only_toydata is True, only generate toydata.

        """
        global_start = time.time()
        cpu_global_start = time.process_time()
        if self.only_toydata:
            self.simulate()
        else:
            results = self.simulate_and_fit()
            self.write_output(results)
        print(
            "Used real time {0:.02f}s, CPU time {1:.02f}s".format(
                time.time() - global_start, time.process_time() - cpu_global_start
            )
        )
        print("Alea iacta est.")



import os
import getpass
import tarfile
import shlex
import json
import tempfile
import time
import threading
import yaml
import subprocess
from alea.submitter import Submitter
import logging
from pathlib import Path
from Pegasus.api import (
    Workflow,
    Job,
    File,
    Site,
    Directory,
    FileServer,
    Operation,
    Namespace,
    Arch,
    Properties,
    SiteCatalog,
    Transformation,
    TransformationCatalog,
    ReplicaCatalog,
)

from datetime import datetime

DEFAULT_IMAGE = "/cvmfs/singularity.opensciencegrid.org/xenonnt/base-environment:latest"
WORK_DIR = f"/scratch/{getpass.getuser()}/workflows"
TOP_DIR = Path(__file__).resolve().parent.parent.parent


# Set up logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger()


class SubmitterHTCondor(Submitter):
    """Submitter for htcondor cluster."""

    def __init__(self, *args, **kwargs):
        # General start
        self.name = self.__class__.__name__
        self.htcondor_configurations = kwargs.get("htcondor_configurations", {})
        self.singularity_image = self.htcondor_configurations.pop(
            "singularity_image", DEFAULT_IMAGE
        )
        self._initial_dir = os.getcwd()
        self.work_dir = WORK_DIR
        self.runs_dir = os.path.join(self.work_dir, "runs")
        self.top_dir = TOP_DIR
        self.template_path = self.htcondor_configurations.pop("template_path", None)
        self.max_jobs_to_combine = self.htcondor_configurations.pop("max_jobs_to_combine", 100)

        # A flag to check if limit_threshold is added to the rc
        self.added_limit_threshold = False

        # Cluster size for toymc jobs
        self.cluster_size = self.htcondor_configurations.pop("cluster_size", 1)

        # Resources configurations
        self.request_cpus = self.htcondor_configurations.pop("request_cpus", 1)
        self.request_memory = self.htcondor_configurations.pop("request_memory", 2000)
        self.request_disk = self.htcondor_configurations.pop("request_disk", 2000000)
        self.combine_disk = self.htcondor_configurations.pop("combine_disk", 20000000)

        # Dagman configurations
        self.dagman_maxidle = self.htcondor_configurations.pop("dagman_maxidle", 100000)
        self.dagman_retry = self.htcondor_configurations.pop("dagman_retry", 2)
        self.dagman_maxjobs = self.htcondor_configurations.pop("dagman_maxjobs", 100000)

        # Pegasus configurations
        self._make_pegasus_config()

        super().__init__(*args, **kwargs)

        # Job input configurations
        self.config_file_path = os.path.abspath(self.config_file_path)
        self._get_statistical_model_config_filename(kwargs)

        # User can provide a name for the workflow, otherwise it will be the current time
        self._setup_wf_id()
        # Pegasus workflow directory
        self.wf_dir = os.path.join(self.runs_dir, self.wf_id)

    def _get_statistical_model_config_filename(self, kwargs):
        """Get the statistical model config file name from the kwargs."""
        _statistical_model_config_filename = kwargs.get("statistical_model_config")
        if os.path.isabs(_statistical_model_config_filename):
            self.statistical_model_config_filename = _statistical_model_config_filename
        else:
            _directory = os.path.dirname(self.config_file_path)
            self.statistical_model_config_filename = os.path.join(
                _directory, _statistical_model_config_filename
            )

    def _validate_x509_proxy(self, min_valid_hours=20):
        """Ensure $X509_USER_PROXY exists and has enough time left.

        This is necessary only if you are going to use Rucio.

        """
        self.x509_user_proxy = os.getenv("X509_USER_PROXY")
        assert self.x509_user_proxy, "Please provide a valid X509_USER_PROXY environment variable."

        logger.debug("Verifying that the X509_USER_PROXY proxy has enough lifetime")
        shell = Shell("grid-proxy-info -timeleft -file %s" % (self.x509_user_proxy))
        shell.run()
        valid_hours = int(shell.get_outerr()) / 60 / 60
        if valid_hours < min_valid_hours:
            raise RuntimeError(
                "User proxy is only valid for %d hours. Minimum required is %d hours."
                % (valid_hours, min_valid_hours)
            )

    def _validate_template_path(self):
        """Validate the template path."""
        assert self.template_path, "Please provide a template path."
        # This path must exists locally, and it will be used to stage the input files
        assert os.path.exists(self.template_path), f"Path {self.template_path} does not exist."

        # Printout the template path file structure
        logger.info("Template path file structure:")
        for dirpath, dirnames, filenames in os.walk(self.template_path):
            for filename in filenames:
                logger.info(f"File: {filename} in {dirpath}")
        if self._contains_subdirectories(self.template_path):
            logger.warning(
                "The template path contains subdirectories. All .h5 files will be tarred."
            )

    def _tar_h5_files(self, directory, output_filename="templates.tar.gz"):
        """Tar all .h5 templates in the directory and its subdirectories into a tarball."""
        # Create a tar.gz archive
        with tarfile.open(output_filename, "w:gz") as tar:
            # Walk through the directory
            for dirpath, dirnames, filenames in os.walk(directory):
                for filename in filenames:
                    if filename.endswith(".h5"):
                        # Get the full path to the file
                        filepath = os.path.join(dirpath, filename)
                        # Add the file to the tar
                        # Specify the arcname to store relative path within the tar
                        tar.add(filepath, arcname=os.path.basename(filename))

    def _make_template_tarball(self):
        """Make tarball of the templates if not exists."""
        self.template_tarball_filename = os.path.join(self._generated_dir(), "templates.tar.gz")
        self._tar_h5_files(self.template_path, self.template_tarball_filename)

    def _modify_yaml(self):
        """Modify the statistical model config file to correct the 'template_filename' fields.

        We will use the modified one to upload to OSG. This modification is necessary because the
        templates on the grid will have different path compared to the local ones, and the
        statistical model config file must reflect that.

        """
        input_file = self.statistical_model_config_filename
        # Output file will have the same name as input file but with '_modified' appended
        _output_file = self._get_file_name(input_file)
        _output_file = _output_file.replace(".yaml", "_modified.yaml")
        output_file = os.path.join(self._generated_dir(), _output_file)
        self.modified_statistical_model_config_filename = output_file

        # Load the YAML data from the original file
        with open(input_file, "r") as file:
            data = yaml.safe_load(file)

        # Recursive function to update 'template_filename' fields
        def update_template_filenames(node):
            if isinstance(node, dict):
                for key, value in node.items():
                    if key == "template_filename":
                        filename = value.split("/")[-1]
                        node[key] = filename
                    else:
                        update_template_filenames(value)
            elif isinstance(node, list):
                for item in node:
                    update_template_filenames(item)

        # Update the data
        update_template_filenames(data)

        # Write the updated YAML data to the new file
        # Overwrite if the file already exists
        with open(output_file, "w") as file:
            yaml.safe_dump(data, file)
        logger.info(f"Modified statistical model config file written to {output_file}")

    def _generated_dir(self, work_dir=WORK_DIR):
        """Directory for generated files."""
        return os.path.join(work_dir, "generated", self.wf_id)

    def _contains_subdirectories(self, directory):
        """Check if the specified directory contains any subdirectories.

        Args:
        directory (str): The path to the directory to check.

        Returns:
        bool: True if there are subdirectories inside the given directory, False otherwise.

        """
        # List all entries in the directory
        try:
            for entry in os.listdir(directory):
                # Check if the entry is a directory
                if os.path.isdir(os.path.join(directory, entry)):
                    return True
        except FileNotFoundError:
            print("The specified directory does not exist.")
            return False
        except PermissionError:
            print("Permission denied for accessing the directory.")
            return False

        # If no subdirectories are found
        return False

    def _setup_wf_id(self):
        """Set up the workflow ID."""
        # If you have named the workflow, use that name. Otherwise, use the current time as name.
        self._wf_id = self.htcondor_configurations.pop("wf_id")
        if self._wf_id:
            self.wf_id = (
                self._wf_id + "-" + self.computation + "-" + datetime.now().strftime("%Y%m%d%H%M")
            )
        else:
            self.wf_id = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    def _pegasus_properties(self):
        """Writes the file pegasus.properties.

        This file contains configuration settings used by Pegasus and HTCondor.

        """
        props = Properties()

        # Don't ask why, but this is necessary for the Pegasus API to work
        props["pegasus.metrics.app"] = "XENON"
        props["pegasus.data.configuration"] = "nonsharedfs"

        # Give jobs a total of 1+{retry} tries
        props["dagman.retry"] = self.dagman_retry
        # Make sure we do start too many jobs at the same time
        props["dagman.maxidle"] = self.dagman_maxidle
        # Total number of jobs cap
        props["dagman.maxjobs"] = self.dagman_maxjobs

        # Help Pegasus developers by sharing performance data
        props["pegasus.monitord.encoding"] = "json"
        props["pegasus.catalog.workflow.amqp.url"] = (
            "amqp://friend:donatedata@msgs.pegasus.isi.edu:5672/prod/workflows"
        )

        # write properties file to ./pegasus.properties
        props.write()
        self.pegasus_properties = props

    def _make_pegasus_config(self):
        """Make the Pegasus configuration into a dict to pass as keywords argument later."""
        pconfig = {}

        pconfig["pegasus.metrics.app"] = "XENON"
        pconfig["pegasus.data.configuration"] = "nonsharedfs"
        pconfig["dagman.retry"] = self.dagman_retry
        pconfig["dagman.maxidle"] = self.dagman_maxidle
        pconfig["dagman.maxjobs"] = self.dagman_maxjobs
        pconfig["pegasus.transfer.threads"] = 4

        # Help Pegasus developers by sharing performance data (optional)
        pconfig["pegasus.monitord.encoding"] = "json"
        pconfig["pegasus.catalog.workflow.amqp.url"] = (
            "amqp://friend:donatedata@msgs.pegasus.isi.edu:5672/prod/workflows"
        )

        self.pegasus_config = pconfig

    def _generate_sc(self):
        """Generates the SiteCatalog for the workflow."""
        sc = SiteCatalog()

        # Local site: this is the submit host
        logger.debug("Defining local site")
        local = Site("local")
        # Logs and pegasus output goes here. This place is called stash in OSG jargon.
        scratch_dir = Directory(
            Directory.SHARED_SCRATCH, path="{}/scratch/{}".format(self.work_dir, self.wf_id)
        )
        scratch_dir.add_file_servers(
            FileServer("file:///{}/scratch/{}".format(self.work_dir, self.wf_id), Operation.ALL)
        )
        # Jobs outputs goes here, but note that it is in scratch so it only stays for short term
        # This place is called stash in OSG jargon.
        storage_dir = Directory(
            Directory.LOCAL_STORAGE, path="{}/outputs/{}".format(self.work_dir, self.wf_id)
        )
        storage_dir.add_file_servers(
            FileServer("file:///{}/outputs/{}".format(self.work_dir, self.wf_id), Operation.ALL)
        )
        # Add scratch and storage directories to the local site
        local.add_directories(scratch_dir, storage_dir)
        # Add profiles to the local site
        local.add_profiles(Namespace.ENV, HOME=os.environ["HOME"])
        local.add_profiles(Namespace.ENV, GLOBUS_LOCATION="")
        local.add_profiles(
            Namespace.ENV,
            PATH=(
                "/cvmfs/xenon.opensciencegrid.org/releases/nT/development/anaconda/envs/XENONnT_development/bin:"  # noqa
                "/cvmfs/xenon.opensciencegrid.org/releases/nT/development/anaconda/condabin:/usr/bin:/bin"  # noqa
            ),
        )
        local.add_profiles(
            Namespace.ENV,
            LD_LIBRARY_PATH=(
                "/cvmfs/xenon.opensciencegrid.org/releases/nT/development/anaconda/envs/XENONnT_development/lib64:"  # noqa
                "/cvmfs/xenon.opensciencegrid.org/releases/nT/development/anaconda/envs/XENONnT_development/lib"  # noqa
            ),
        )
        local.add_profiles(Namespace.ENV, PEGASUS_SUBMITTING_USER=os.environ["USER"])
        local.add_profiles(Namespace.ENV, X509_USER_PROXY=os.environ["X509_USER_PROXY"])

        # Staging sites: for XENON it is physically at dCache in UChicago
        # You will be able to download results from there via gfal commands
        logger.debug("Defining stagging site")
        staging_davs = Site("staging-davs")
        scratch_dir = Directory(
            Directory.SHARED_SCRATCH, path="/xenon/scratch/{}".format(getpass.getuser())
        )
        scratch_dir.add_file_servers(
            FileServer(
                "gsidavs://xenon-gridftp.grid.uchicago.edu:2880/xenon/scratch/{}".format(
                    getpass.getuser()
                ),
                Operation.ALL,
            )
        )
        staging_davs.add_directories(scratch_dir)

        # Condorpool: These are the job nodes on grid
        logger.debug("Defining condorpool")
        condorpool = Site("condorpool")
        condorpool.add_profiles(Namespace.PEGASUS, style="condor")
        condorpool.add_profiles(Namespace.CONDOR, universe="vanilla")
        condorpool.add_profiles(
            Namespace.CONDOR, key="+SingularityImage", value=f'"{self.singularity_image}"'
        )
        # Ignore the site settings, since the container will set all this up inside
        condorpool.add_profiles(Namespace.ENV, OSG_LOCATION="")
        condorpool.add_profiles(Namespace.ENV, GLOBUS_LOCATION="")
        condorpool.add_profiles(Namespace.ENV, PYTHONPATH="")
        condorpool.add_profiles(Namespace.ENV, PERL5LIB="")
        condorpool.add_profiles(Namespace.ENV, LD_LIBRARY_PATH="")
        condorpool.add_profiles(Namespace.ENV, PEGASUS_SUBMITTING_USER=os.environ["USER"])
        condorpool.add_profiles(
            Namespace.CONDOR, key="x509userproxy", value=os.environ["X509_USER_PROXY"]
        )

        # Add the sites to the SiteCatalog
        sc.add_sites(local, staging_davs, condorpool)
        return sc

    def _generate_tc(self, cluster_size=1):
        """Generates the TransformationCatalog for the workflow.

        Every executable that is used in the workflow should be here.

        """
        # Wrappers that runs alea-run_toymc
        run_toymc_wrapper = Transformation(
            name="run_toymc_wrapper",
            site="local",
            pfn=self.top_dir / "alea/submitters/run_toymc_wrapper.sh",
            is_stageable=True,
            arch=Arch.X86_64,
        ).add_pegasus_profile(clusters_size=cluster_size)

        # Wrappers that collect outputs
        combine = Transformation(
            name="combine",
            site="local",
            pfn=self.top_dir / "alea/submitters/combine.sh",
            is_stageable=True,
            arch=Arch.X86_64,
        )

        tc = TransformationCatalog()
        tc.add_transformations(run_toymc_wrapper, combine)

        return tc

    def _generate_rc(self):
        """Generate the ReplicaCatalog for the workflow.

        1. The input files for the job, which are the templates in tarball,
            the yaml files and alea-run_toymc.
        2. The output files for the job, which are the toydata and the output files.
        Since the outputs are not known in advance, we will add them in the job definition.

        """
        rc = ReplicaCatalog()

        # Add the templates
        self.f_template_tarball = File(str(self._get_file_name(self.template_tarball_filename)))
        rc.add_replica(
            "local",
            str(self._get_file_name(self.template_tarball_filename)),
            "file://{}".format(self.template_tarball_filename),
        )
        # Add the yaml files
        self.f_running_configuration = File(str(self._get_file_name(self.config_file_path)))
        rc.add_replica(
            "local",
            str(self._get_file_name(self.config_file_path)),
            "file://{}".format(self.config_file_path),
        )
        self.f_statistical_model_config = File(
            str(self._get_file_name(self.modified_statistical_model_config_filename))
        )
        rc.add_replica(
            "local",
            str(self._get_file_name(self.modified_statistical_model_config_filename)),
            "file://{}".format(self.modified_statistical_model_config_filename),
        )
        # Add run_toymc_wrapper
        self.f_run_toymc_wrapper = File("run_toymc_wrapper.sh")
        rc.add_replica(
            "local",
            "run_toymc_wrapper.sh",
            "file://{}".format(self.top_dir / "alea/submitters/run_toymc_wrapper.sh"),
        )
        # Add alea-run_toymc
        self.f_alea_run_toymc = File("alea-run_toymc")
        rc.add_replica(
            "local",
            "alea-run_toymc",
            "file://{}".format(self.top_dir / "bin/alea-run_toymc"),
        )
        #  Add combine executable
        self.f_combine = File("combine.sh")
        rc.add_replica(
            "local",
            "combine.sh",
            "file://{}".format(self.top_dir / "alea/submitters/combine.sh"),
        )

        return rc

    def _add_combine_job(self, combine_i):
        """Add a combine job to the workflow."""
        logger.info(f"Adding combine job {combine_i} to the workflow")
        combine_name = "combine"
        combine_job = self._initialize_job(
            name=combine_name,
            cores=self.request_cpus,
            memory=self.request_memory * 2,
            disk=self.combine_disk,
        )
        combine_job.add_profiles(Namespace.CONDOR, "requirements", self.requirements)

        # Combine job configuration: all toymc results and files will be combined into one tarball
        combine_job.add_outputs(
            File("%s-%s-combined_output.tar.gz" % (self.wf_id, combine_i)), stage_out=True
        )
        combine_job.add_args(self.wf_id + "-" + str(combine_i))
        self.wf.add_jobs(combine_job)

        return combine_job

    def _generate_workflow(self, name="run_toymc_wrapper"):
        """Generate the workflow.

        1. Define catalogs
        2. Generate jobs by iterating over the path-modified tickets
        3. Add jobs to the workflow

        """
        # Initialize the workflow
        self.wf = Workflow("alea-workflow")
        self.sc = self._generate_sc()
        self.tc = self._generate_tc(cluster_size=self.cluster_size)
        self.rc = self._generate_rc()

        # Job requirements
        self.requirements = self._make_requirements()

        # Iterate over the tickets and generate jobs
        combine_i = 0
        new_to_combine = True
        # Generate jobstring and output names from tickets generator
        # _script for example:
        # alea-submission lq_b8_cevns_running.yaml --computation discovery_power --local --debug
        # _last_output_filename for example: /project/lgrandi/yuanlq/alea_outputs/b8mini/toymc_power_cevns_livetime_1.22_0.50_b8_rate_1.00_0.h5  # noqa
        # _script for example: python3 /home/yuanlq/.local/bin/alea-run_toymc --statistical_model alea.models.BlueiceExtendedModel --poi b8_rate_multiplier --hypotheses '["free","zero","true"]' --n_mc 50 --common_hypothesis None --generate_values '{"b8_rate_multiplier":1.0}' --nominal_values '{"livetime_sr0":1.221,"livetime_sr1":0.5}' --statistical_model_config lq_b8_cevns_statistical_model.yaml --parameter_definition None --statistical_model_args '{"template_path":"/project2/lgrandi/binference_common/nt_cevns_templates"}' --likelihood_config None --compute_confidence_interval False --confidence_level 0.9000 --confidence_interval_kind central --toydata_mode generate_and_store --toydata_filename /project/lgrandi/yuanlq/alea_outputs/b8mini/toyfile_cevns_livetime_1.22_0.50_b8_rate_1.00_0.h5 --only_toydata False --output_filename /project/lgrandi/yuanlq/alea_outputs/b8mini/toymc_power_cevns_livetime_1.22_0.50_b8_rate_1.00_0.h5 --seed None --metadata None  # noqa
        for jobid, (_script, _) in enumerate(self.combined_tickets_generator()):
            # If the number of jobs to combine is reached, add a new combine job
            if new_to_combine:
                combine_job = self._add_combine_job(combine_i)

            # Reorganize the script to get the executable and arguments,
            # in which the paths are corrected
            executable, args_dict = self._reorganize_script(_script)
            if not (args_dict["toydata_mode"] in ["generate_and_store", "generate"]):
                raise NotImplementedError(
                    "Only generate_and_store toydata mode is supported on OSG."
                )

            logger.info(f"Adding job {jobid} to the workflow")
            logger.debug(f"Naked Script: {_script}")
            logger.debug(f"Output: {args_dict['output_filename']}")
            logger.debug(f"Executable: {executable}")
            logger.debug(f"Toydata: {args_dict['toydata_filename']}")
            logger.debug(f"Arguments: {args_dict}")

            # Create a job with base requirements
            job = self._initialize_job(
                name=name,
                cores=self.request_cpus,
                memory=self.request_memory,
                disk=self.request_disk,
            )
            job.add_profiles(Namespace.CONDOR, "requirements", self.requirements)

            # Add the inputs and outputs
            job.add_inputs(
                self.f_template_tarball,
                self.f_running_configuration,
                self.f_statistical_model_config,
                self.f_run_toymc_wrapper,
                self.f_alea_run_toymc,
            )
            if self.added_limit_threshold:
                job.add_inputs(self.f_limit_threshold)

            job.add_outputs(File(args_dict["output_filename"]), stage_out=False)
            combine_job.add_inputs(File(args_dict["output_filename"]))

            # Only add the toydata file if instructed to do so
            if args_dict["toydata_mode"] == "generate_and_store":
                job.add_outputs(File(args_dict["toydata_filename"]), stage_out=False)
                combine_job.add_inputs(File(args_dict["toydata_filename"]))

            # Add the arguments into the job
            # Using escaped argument to avoid the shell syntax error
            def _extract_all_to_tuple(d):
                return tuple(
                    f"{json.dumps(str(d[key])).replace(' ', '')}".replace("'", '\\"')
                    for key in d.keys()
                )

            args_tuple = _extract_all_to_tuple(args_dict)
            job.add_args(*args_tuple)

            # Add the job to the workflow
            self.wf.add_jobs(job)

            # If the number of jobs to combine is reached, add a new combine job
            if (jobid + 1) % self.max_jobs_to_combine == 0:
                new_to_combine = True
                combine_i += 1
            else:
                new_to_combine = False

        # Finalize the workflow
        os.chdir(self._generated_dir())
        self.wf.add_replica_catalog(self.rc)
        self.wf.add_transformation_catalog(self.tc)
        self.wf.add_site_catalog(self.sc)
        self.wf.write()

    def _initialize_job(
        self,
        name="run_toymc_wrapper",
        run_on_submit_node=False,
        cores=1,
        memory=1_700,
        disk=1_000_000,
    ):
        """Initilize a Pegasus job, also sets resource profiles.

        Memory in unit of MB, and disk in unit of MB.

        """
        job = Job(name)
        job.add_profiles(Namespace.CONDOR, "request_cpus", str(cores))

        if run_on_submit_node:
            job.add_selector_profile(execution_site="local")
            # no other attributes on a local job
            return job

        # Set memory and disk requirements
        # If the job fails, retry with more memory and disk
        memory_str = (
            "ifthenelse(isundefined(DAGNodeRetry) || "
            f"DAGNodeRetry == 0, {memory}, (DAGNodeRetry + 1)*{memory})"
        )
        disk_str = (
            "ifthenelse(isundefined(DAGNodeRetry) || "
            f"DAGNodeRetry == 0, {disk}, (DAGNodeRetry + 1)*{disk})"
        )
        job.add_profiles(Namespace.CONDOR, "request_disk", disk_str)
        job.add_profiles(Namespace.CONDOR, "request_memory", memory_str)

        return job

    def _make_requirements(self):
        """Make the requirements for the job."""
        # Minimal requirements on singularity/cvmfs/ports/microarchitecture
        requirements_base = (
            "HAS_SINGULARITY && HAS_CVMFS_xenon_opensciencegrid_org"
            + " && PORT_2880 && PORT_8000 && PORT_27017"
            + ' && (Microarch >= "x86_64-v3")'
        )

        # If in debug mode, use the MWT2 site because we own it
        if self.debug:
            requirements_base = requirements_base + ' && GLIDEIN_ResourceName == "MWT2" '

        requirements = requirements_base
        return requirements

    def _reorganize_script(self, _script):
        """Extract executable and arguments from the naked scripts.

        Correct the paths on the fly.

        """
        script_fragments = _script.split(" ")
        _executable = script_fragments[1]
        # Remove the directory and only keep the file name
        executable = self._get_file_name(_executable)

        args_dict = self._parse_command_args(_script)

        # Add the limit_threshold to the replica catalog if not added
        if (
            not self.added_limit_threshold
            and "limit_threshold" in args_dict["statistical_model_args"].keys()
        ):
            self.limit_threshold_filename = args_dict["statistical_model_args"]["limit_threshold"]
            self._add_limit_threshold()

        # Correct the paths in the arguments
        args_dict = self._correct_paths_args_dict(args_dict)

        return executable, args_dict

    def _add_limit_threshold(self):
        """Add the Neyman thresholds limit_threshold to the replica catalog."""
        self.f_limit_threshold = File(str(self._get_file_name(self.limit_threshold_filename)))
        self.rc.add_replica(
            "local",
            str(self._get_file_name(self.limit_threshold_filename)),
            "file://{}".format(self.limit_threshold_filename),
        )
        self.added_limit_threshold = True

    def _correct_paths_args_dict(self, args_dict):
        """Correct the paths in the arguments dictionary in a hardcoding way."""
        args_dict["statistical_model_args"]["template_path"] = "templates/"

        if "limit_threshold" in args_dict["statistical_model_args"].keys():
            limit_threshold_filename = self._get_file_name(
                args_dict["statistical_model_args"]["limit_threshold"]
            )
            args_dict["statistical_model_args"]["limit_threshold"] = limit_threshold_filename

        toydata_filename = self._get_file_name(args_dict["toydata_filename"])
        args_dict["toydata_filename"] = toydata_filename

        output_filename = self._get_file_name(args_dict["output_filename"])
        args_dict["output_filename"] = output_filename

        statistical_model_config = self._get_file_name(
            self.modified_statistical_model_config_filename
        )
        args_dict["statistical_model_config"] = statistical_model_config

        return args_dict

    def _parse_command_args(self, command):
        """Parse the command line arguments and return a dictionary with the flags and their values.

        Example command: python3 /home/yuanlq/.local/bin/alea-run_toymc --statistical_model alea.models.BlueiceExtendedModel --poi b8_rate_multiplier --hypotheses '["free","zero","true"]' --n_mc 50 --common_hypothesis None --generate_values '{"b8_rate_multiplier":1.0}' --nominal_values '{"livetime_sr0":1.221,"livetime_sr1":0.5}' --statistical_model_config lq_b8_cevns_statistical_model.yaml --parameter_definition None --statistical_model_args '{"template_path":"/project2/lgrandi/binference_common/nt_cevns_templates"}' --likelihood_config None --compute_confidence_interval False --confidence_level 0.9000 --confidence_interval_kind central --toydata_mode generate_and_store --toydata_filename /project/lgrandi/yuanlq/alea_outputs/b8mini/toyfile_cevns_livetime_1.22_0.50_b8_rate_1.00_0.h5 --only_toydata False --output_filename /project/lgrandi/yuanlq/alea_outputs/b8mini/toymc_power_cevns_livetime_1.22_0.50_b8_rate_1.00_0.h5 --seed None --metadata None  # noqa

        """
        # Use shlex to handle spaces within quotes correctly
        parts = shlex.split(command)

        # Dictionary to store command flags and their values
        args_dict = {}

        # Iterator to go through all parts
        it = iter(parts)
        for part in it:
            if part.startswith("--"):
                flag = part[2:]  # Remove '--' from the flag name
                next_part = next(it, None)
                # Check if the next part is also a flag or end of the command
                if next_part is None or next_part.startswith("--"):
                    args_dict[flag] = True  # No value means boolean flag set to True
                    continue
                # Try to handle JSON values correctly
                if next_part.startswith("{"):
                    # Reconstruct the complete JSON by consuming parts until the closing '}'
                    json_value = next_part
                    while not json_value.strip().endswith("}"):
                        json_value += " " + next(it, "")
                    try:
                        args_dict[flag] = json.loads(json_value)
                    except json.JSONDecodeError:
                        args_dict[flag] = json_value  # Store raw string if JSON parsing fails
                else:
                    args_dict[flag] = next_part

        return args_dict

    def _get_file_name(self, file_path):
        """Get the filename from the file path."""
        return os.path.basename(file_path)

    def _us_sites_only(self):
        raise NotImplementedError

    def _exclude_sites(self):
        raise NotImplementedError

    def _this_site_only(self):
        raise NotImplementedError

    def _check_workflow_exists(self):
        """Check if the workflow already exists."""
        if os.path.exists(self.wf_dir):
            logger.error(f"Workflow already exists at {self.wf_dir}. Exiting.")
            return

    def _plan_and_submit(self):
        """Plan and submit the workflow."""
        os.chdir(self._generated_dir())
        self.wf.plan(
            submit=not self.debug,
            cluster=["horizontal"],
            cleanup="none",
            sites=["condorpool"],
            verbose=3 if self.debug else 0,
            staging_sites={"condorpool": "staging-davs"},
            output_sites=["local"],
            dir=os.path.dirname(self.wf_dir),
            relative_dir=self.wf_id,
            **self.pegasus_config,
        )

        print(f"Worfklow written to \n\n\t{self.wf_dir}\n\n")

    def _warn_outputfolder(self):
        """Warn users about the outputfolder in running config won't be really used."""
        logger.warning(
            "The outputfolder in the running configuration %s won't be used in this submission."
            % (self.outputfolder)
        )
        logger.warning(
            "Instead, you should find your outputs at %s"
            % (self.work_dir + "/outputs/" + self.wf_id)
        )

    def _check_filename_unique(self):
        """Check if all the files in the template path are unique."""
        all_files = []
        for _, _, filenames in os.walk(self.template_path):
            for filename in filenames:
                all_files.append(filename)
        if len(all_files) != len(set(all_files)):
            raise RuntimeError("All files in the template path must have unique names.")

    def submit(self, **kwargs):
        """Serve as the main function to submit the workflow."""
        self._check_workflow_exists()
        self._validate_x509_proxy()

        #  0o755 means read/write/execute for owner, read/execute for everyone else
        try:
            os.makedirs(self._generated_dir(), 0o755)
        except FileExistsError:
            logger.error(f"Workflow directory {self._generated_dir()} already exists. Exiting.")
        os.makedirs(self.runs_dir, 0o755, exist_ok=True)

        # Modify the statistical model config file to correct the 'template_filename' fields
        self._modify_yaml()

        # Handling templates as part of the inputs
        self._validate_template_path()
        self._check_filename_unique()
        self._make_template_tarball()

        self._generate_workflow()
        self._plan_and_submit()

        # Return to initial dir, as we are done.
        logger.info("We are done. Returning to initial directory.")
        self._warn_outputfolder()
        os.chdir(self._initial_dir)


class Shell(object):
    """Provides a shell callout with buffered stdout/stderr, error handling and timeout."""

    def __init__(self, cmd, timeout_secs=1 * 60 * 60, log_cmd=False, log_outerr=False):
        self._cmd = cmd
        self._timeout_secs = timeout_secs
        self._log_cmd = log_cmd
        self._log_outerr = log_outerr
        self._process = None
        self._out_file = None
        self._outerr = ""
        self._duration = 0.0

    def run(self):
        def target():

            self._process = subprocess.Popen(
                self._cmd,
                shell=True,
                stdout=self._out_file,
                stderr=subprocess.STDOUT,
                preexec_fn=os.setpgrp,
            )
            self._process.communicate()

        if self._log_cmd:
            print(self._cmd)

        # temp file for the stdout/stderr
        self._out_file = tempfile.TemporaryFile(prefix="outsource-", suffix=".out")

        ts_start = time.time()

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(self._timeout_secs)
        if thread.is_alive():
            # do our best to kill the whole process group
            try:
                kill_cmd = "kill -TERM -%d" % (os.getpgid(self._process.pid))
                kp = subprocess.Popen(kill_cmd, shell=True)
                kp.communicate()
                self._process.terminate()
            except Exception:
                pass
            thread.join()
            # log the output
            self._out_file.seek(0)
            stdout = self._out_file.read().decode("utf-8").strip()
            if self._log_outerr and len(stdout) > 0:
                print(stdout)
            self._out_file.close()
            raise RuntimeError(
                "Command timed out after %d seconds: %s" % (self._timeout_secs, self._cmd)
            )

        self._duration = time.time() - ts_start

        # log the output
        self._out_file.seek(0)
        self._outerr = self._out_file.read().decode("utf-8").strip()
        if self._log_outerr and len(self._outerr) > 0:
            print(self._outerr)
        self._out_file.close()

        if self._process.returncode != 0:
            raise RuntimeError(
                "Command exited with non-zero exit code (%d): %s\n%s"
                % (self._process.returncode, self._cmd, self._outerr)
            )

    def get_outerr(self):
        """Returns the combined stdout and stderr from the command."""
        return self._outerr

    def get_exit_code(self):
        """Returns the exit code from the process."""
        return self._process.returncode

    def get_duration(self):
        """Returns the timing of the command (seconds)"""
        return self._duration
